/*  Termite filter: auto-reply to incoming text that matches a pattern
 *
 *  Copyright (c) CompuPhase, 2015-2018
 *
 *  This software is provided "as-is", without any express or implied warranty.
 *  In no event will the authors be held liable for any damages arising from
 *  the use of this software.
 *
 *  Permission is granted to anyone to use this software for any purpose,
 *  including commercial applications, and to alter it and redistribute it
 *  freely, subject to the following restrictions:
 *
 *  1.  The origin of this software must not be misrepresented; you must not
 *      claim that you wrote the original software. If you use this software in
 *      a product, an acknowledgment in the product documentation would be
 *      appreciated but is not required.
 *  2.  Altered source versions must be plainly marked as such, and must not be
 *      misrepresented as being the original software.
 *  3.  This notice may not be removed or altered from any source distribution.
 *
 *  Nota Bene:
 *  This software uses the "Super Light Regular Expression" library by Sergey
 *  Lyubka. This library is distributed under the terms of the GNU General
 *  Public License version 2.
 *
 *  That is: the source code of the Termite "Auto-Reply" filter is subject to
 *  the *  zlib/libpng license (as mentioned above) and the SLRE source code is
 *  subject to the GPL version 2.
 */

#include <assert.h>
#include <windows.h>
#include <string.h>
#include "slre.h"
#include "textbuffer.h"
#include "autoreply.h"


#define FLTEXPORT __declspec(dllexport) __stdcall

#if !defined sizearray
  #define sizearray(a)  (sizeof(a) / sizeof((a)[0]))
#endif

    /* To create a window in Termites interface, the plugin must send the
     * message UM_PLUGINWINDOW to the Termite main window from its flt_Load()
     * function. That is, the plugin must have an flt_Load() function and it
     * must call "SendMessage(hwnd,UM_PLUGINWINDOW,nnn,0L);"
     * The wParam parameter ("nnn" in the above example) is the height of the
     * window (in pixels). The return value of the SendMessage() call is the
     * window handle. Typically, a plugin will subclass this window in order to
     * receive notifications for any controls that it creates in it.
     */
#define UM_PLUGINWINDOW (WM_APP + 1)  /* also defined in filter.h */

    /* Originally intended to transmit a data buffer on a hotkey, this message
     * can be used to transmit a block of data. The lParam parameter must be
     * set to point to the data buffer and the wParam parameter must be set to
     * the size of the buffer.
     */
#define UM_POSTHOTKEY   (WM_APP + 3)  /* also defined in filter.h */

    /* a few strings that are used in various locations */
const char strCaption[] = "Auto-Reply filter";
const char strSection[] = "AutoReply";
const char strExprKey[] = "match%d";
const char strReplyKey[] = "reply%d";

    /* forward declarations of a few functions */
LRESULT CALLBACK SubClassFunc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK FilterConfigDlg(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK FilterInfoDlg(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
static void reload(BOOL clearvars);
static BOOL parse_var(const char *string, char *label, char *defvalue, int *vpos, int *vlen);
static int match_re(const char *string, int length, int *scanned);
static int match_hex(const char *string, int length, int *scanned);
static BOOL strparse(char *output, int size, const char *input);
static int string2hex(const char *string, unsigned char *buffer, int buflen);
static char *strdel(char *str,size_t len);
static char *strins(char *dest,const char *src,size_t srclen);


    /* this variable is declared in dllmain.c */
extern HINSTANCE hinstDLL;

    /* a few global types and variables for this particular filter */
#define NUM_EXPRESSIONS   5
#define MAX_EXPR_LEN      128
typedef struct tagEXPR {
  char match[MAX_EXPR_LEN];
  char reply[MAX_EXPR_LEN];
  int match_len;
  int reply_len;
  BOOL casesensitive;
  BOOL hexmode;
} EXPR;
static EXPR Expression[NUM_EXPRESSIONS];

#define NUM_VARIABLES     5
#define MAX_VARLABEL_LEN  16
#define MAX_VARDATA_LEN   16
typedef struct tagVARIABLE {
  HWND hwndLabel;         /* handle to the label ("static" control) */
  HWND hwndEdit;          /* handle to the edit field */
  int expr_id;            /* index in the "Expression" array */
  int seq;                /* sequence, if there are multiple variables in an expression */
  char label[MAX_VARLABEL_LEN];
  char data[MAX_VARDATA_LEN];
} VARIABLE;
static VARIABLE Variable[NUM_EXPRESSIONS];

static int BufferSize = 0;
static int BufferRemove = 0;

static HWND hwndTermite;  /* handle of Termite's main window */
static HWND hwndPlugin;   /* handle of the sub-window that Termite creates for this plugin */
static char ProfileFile[_MAX_PATH];
static HFONT hfontNormal;
static HBITMAP hbmpHelp;
static WNDPROC lpfnOldWndProc;


/* flt_Load() is called when Termite loads the filter.
 *
 * hwnd         The handle to the main window of Termite.
 * ProfileName  The full path to the INI file that Termite uses. The filter
 *              can use this name to store its configuration.
 * Build        The build number of Termite, for distinguishing versions of
 *              the Termite application.
 *
 * This function must return TRUE if it can load successfully. If the function
 * returns FALSE, Termite will unload it.
 */
BOOL FLTEXPORT flt_Load(HWND hwnd, LPCSTR ProfileName, int Build)
{
  int count, idx;

  (void)Build;
  hwndTermite = hwnd;
  hwndPlugin = NULL;
  strcpy(ProfileFile, ProfileName); /* save name of the INI file */

  BufferSize = 0;
  BufferRemove = 0;

  reload(TRUE);

  /* check whether there are variables */
  for (count = idx = 0; idx < NUM_VARIABLES; idx++)
    if (strlen(Variable[idx].label) > 0 && strlen(Variable[idx].data) > 0)
      count += 1;

  if (count > 0) {
    hwndPlugin = (HWND)SendMessage(hwndTermite, UM_PLUGINWINDOW, 18, 0L);
    if (hwndPlugin) {
      hfontNormal = CreateFont(14, 0, 0, 0, FW_NORMAL, FALSE, FALSE, FALSE,
                             ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
                             DEFAULT_QUALITY, FF_DONTCARE | DEFAULT_PITCH,
                             "Arial");

      /* create help button */
      CreateWindow("button", "HELP_BUTTON", WS_CHILD | WS_VISIBLE | BS_BITMAP | BS_PUSHBUTTON | BS_FLAT, 0, 0, 17, 17, hwndPlugin, (HMENU)ID_HELP, hinstDLL, NULL);
      hbmpHelp = LoadImage(hinstDLL, "HELP_BUTTON", IMAGE_BITMAP, 16, 16, LR_DEFAULTSIZE);
      SendDlgItemMessage(hwndPlugin, ID_HELP, BM_SETIMAGE, 0, (LPARAM)hbmpHelp);

      /* create control with caption */
      CreateWindow("static", "LOGO_IMG", WS_CHILD | WS_VISIBLE | SS_BITMAP, 24, 1, 60, 16, hwndPlugin, (HMENU)(-1), hinstDLL, NULL);

      /* subclass the window to intercept the notification messages of the controls */
      lpfnOldWndProc = (WNDPROC)SetWindowLong(hwndPlugin, GWL_WNDPROC, (DWORD)SubClassFunc);

      /* reload the definitions again, because the controls can only becreated
         after this toolbar window is made */
      reload(TRUE);
    }
  }

  return TRUE;
}


/* flt_Unload() is called when Termite removes the filter.
 *
 * After calling this notification function, Termite will physically unload
 * the filter DLL.
 *
 * For filters that allocate dynamic memory or other resources, this is a good
 * moment to free these resources.
 */
void FLTEXPORT flt_Unload(void)
{
  /* free text buffer */
  txtbuf_Free();
}


/* flt_Flags() is called when Termite starts up, and browses through all
 * filters to verify their class and settings.
 */
int FLTEXPORT flt_Flags(void)
{
  return 0x04 | 0x08; /* tell Termite that this filter must process the data
                         first (priority) but that it should not be called on
                         local echo */
}


/* flt_Receive() is called after Termite has received new data. It is also
 * called after not receiving data for some time-out (which is hard-coded to
 * 0.5 second). If a plug-in buffers data internally, this time-out allows the
 * plug-in to parse the remaining data.
 *
 * Text         The contents of the received data.
 * Size         On input, this parameter points to the size of the data block
 *              that parameter "Text" holds. Note that the data in "Text" need
 *              to be zero-terminated. On output, this parameter must hold the
 *              new size of the data block.
 *              This parameter may be zero on input, so that the plug-in can
 *              pass any data that it had buffered internally to Termite.
 *
 * If flt_Receive() does not change the string, it can return NULL or return
 * the original string.
 *
 * If the result of flt_Receive() is a shorter string than the input string,
 * the function may change the string in place (but this is not encouraged).
 * If the string changes (and especially if the result is bigger than the input
 * string), the function should allocate memory for the modified string. The
 * function should also keep track of the allocated memory, because it must
 * free the memory itself: either on a next call, or on flt_Unload(). It is
 * suggested that the filter creates an auto-growing output buffer.
 *
 * If the filter wishes to remove all data, it must set parameter "Size" to
 * zero on output. It should return a pointer to the input buffer (parameter
 * "Text"); specifically, it should not return NULL.
 *
 * The input and output strings are not necessarily zero-terminated; the
 * filter must adjust the length to the number of bytes it returns (parameter
 * "Size").
 */
LPCSTR FLTEXPORT flt_Receive(LPCSTR Text, LPINT Size)
{
  int numbytes;
  int matchidx, matchoffs, matchscanned;
  LPSTR BufPtr, Ptr, Tail;

  /* see whether there is data in the buffer that should be stripped (handled
     from the previous call) */
  if (BufferRemove > 0) {
    BufPtr = txtbuf_Alloc(BufferSize);  /* just returns the current buffer (no resize) */
    assert(BufPtr != NULL);
    assert(BufferRemove <= BufferSize);
    if (BufferRemove < BufferSize)
      memmove(BufPtr, BufPtr + BufferRemove, BufferSize - BufferRemove);
    BufferSize -= BufferRemove;
    assert(BufferSize >= 0);
    BufferRemove = 0;
  } /* if */

  /* buffer a complete line */
  numbytes = BufferSize + *Size;
  BufPtr = txtbuf_Alloc(numbytes);
  if (BufPtr == NULL)
    return NULL;  /* this can also happen if both BufferSize & *Size are zero */

  if (*Size > 0) {
    memcpy(BufPtr + BufferSize, Text, *Size);
    BufferSize = numbytes;

    /* hex-mode match */
    matchoffs = 0;
    for ( ;; ) {
      matchidx = match_hex(BufPtr + matchoffs, BufferSize - matchoffs, &matchscanned);
      if (matchidx < 0)
        break;
      PostMessage(hwndTermite, UM_POSTHOTKEY, Expression[matchidx].reply_len, (long)Expression[matchidx].reply);
      matchoffs += matchscanned;
    }

    /* for regex match see whether we have at least one complete line in the buffer */
    Tail = NULL;
    Ptr = BufPtr;
    for ( ;; ) {
      while ((Ptr - BufPtr) < numbytes && *Ptr != '\r' && *Ptr != '\n')
        Ptr++;
      if ((Ptr - BufPtr) >= numbytes)
        break;
      while ((Ptr - BufPtr) < numbytes && (*Ptr == '\r' || *Ptr == '\n'))
        Ptr++;
      Tail = Ptr;
    } /* for */
    /* if no line, gobble all (unless there is recognized hex data) */
    if (Tail == NULL) {
      BufferRemove = matchoffs;
      *Size = matchoffs;
      return (matchoffs == 0) ? NULL : BufPtr;
    } /* if */
    /* if arrived here, there is at least one full line in the buffer (possibly more) */
    assert(Tail != NULL);
    BufferRemove = (int)(Tail - BufPtr);
    matchoffs = 0;
    for ( ;; ) {
      matchidx = match_re(BufPtr + matchoffs, BufferRemove - matchoffs, &matchscanned);
      if (matchidx < 0)
        break;
      PostMessage(hwndTermite, UM_POSTHOTKEY, Expression[matchidx].reply_len, (long)Expression[matchidx].reply);
      matchoffs += matchscanned;
    }
    *Size = BufferRemove;
  } else {
    /* special case, if *Size is zero on input, return the buffered data */
    assert(*Size == 0);
    BufferRemove = BufferSize;
    matchoffs = 0;
    for ( ;; ) {
      matchidx = match_hex(BufPtr + matchoffs, BufferRemove - matchoffs, &matchscanned);
      if (matchidx < 0)
        matchidx = match_re(BufPtr + matchoffs, BufferRemove - matchoffs, &matchscanned);
      if (matchidx < 0)
        break;
      PostMessage(hwndTermite, UM_POSTHOTKEY, Expression[matchidx].reply_len, (long)Expression[matchidx].reply);
      matchoffs += matchscanned;
    }
    *Size = BufferSize;
    BufferSize = 0;
    BufferRemove = 0;
  } /* if */

  return BufPtr;
}


BOOL FLTEXPORT flt_Config(void)
{
  /* pop up the dialog for configuration */
  int result = DialogBox(hinstDLL, "FilterConfig", GetFocus(), (DLGPROC)FilterConfigDlg);
  if (result == IDOK)
    reload(TRUE);

  return (result == IDOK);
}

#pragma argsused
BOOL CALLBACK FilterConfigDlg(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
  int idx;
  BOOL hexmode;
  char str[2*MAX_EXPR_LEN], key[32], *ptr;

  switch (msg) {
  case WM_INITDIALOG:
    /* read the settings from the INI file */
    for (idx = 0; idx < NUM_EXPRESSIONS; idx++) {
      wsprintf(key, strExprKey, idx + 1);
      GetPrivateProfileString(strSection, key, "", str, sizearray(str), ProfileFile);
      ptr = strchr(str, '/');
      if (ptr != NULL && strlen(ptr + 1) < MAX_EXPR_LEN) {
        *ptr++ = '\0';  /* skip '/' */
        SetDlgItemText(hwnd, IDD_MATCH1 + idx, ptr);
        SendDlgItemMessage(hwnd, IDD_MATCH1 + idx, EM_LIMITTEXT, MAX_EXPR_LEN - 1, 0L);
        hexmode = (strchr(str, 'x') != NULL);
        CheckDlgButton(hwnd, IDD_CASE1 + idx, (strchr(str, 'i') == NULL) ? BST_CHECKED : BST_UNCHECKED);
        CheckDlgButton(hwnd, IDD_HEX1 + idx, hexmode ? BST_CHECKED : BST_UNCHECKED);
        wsprintf(key, strReplyKey, idx + 1);
        GetPrivateProfileString(strSection, key, "", str, sizearray(str), ProfileFile);
        SetDlgItemText(hwnd, IDD_REPLY1 + idx, str);
        SendDlgItemMessage(hwnd, IDD_REPLY1 + idx, EM_LIMITTEXT, MAX_EXPR_LEN - 1, 0L);
        EnableWindow(GetDlgItem(hwnd, IDD_CASE1 + idx), !hexmode);
      } else {
        CheckDlgButton(hwnd, IDD_CASE1 + idx, BST_CHECKED);
      }
    } /* for */
    return TRUE;

  case WM_COMMAND:
    switch (LOWORD(wParam)) {
    case IDCANCEL:
      EndDialog(hwnd, LOWORD(wParam));
      return TRUE;
    case IDOK:
      /* save settings into the INI file */
      for (idx = 0; idx < NUM_EXPRESSIONS; idx++) {
        str[0] = '\0';
        if (IsDlgButtonChecked(hwnd, IDD_CASE1 + idx) != BST_CHECKED)
          strcat(str, "i");
        if (IsDlgButtonChecked(hwnd, IDD_HEX1 + idx) == BST_CHECKED)
          strcat(str, "x");
        strcat(str, "/");
        GetDlgItemText(hwnd, IDD_MATCH1 + idx, str + strlen(str), MAX_EXPR_LEN);
        wsprintf(key, strExprKey, idx + 1);
        WritePrivateProfileString(strSection, key, str, ProfileFile);
        GetDlgItemText(hwnd, IDD_REPLY1 + idx, str, MAX_EXPR_LEN);
        wsprintf(key, strReplyKey, idx + 1);
        WritePrivateProfileString(strSection, key, str, ProfileFile);
      }
      EndDialog(hwnd, LOWORD(wParam));
      return TRUE;
    case IDHELP:
      DialogBox(hinstDLL, "FilterInfo", hwnd, (DLGPROC)FilterInfoDlg);
      return TRUE;
    case IDD_HEX1:
    case IDD_HEX2:
    case IDD_HEX3:
    case IDD_HEX4:
    case IDD_HEX5:
    case IDD_HEX6:
    case IDD_HEX7:
    case IDD_HEX8:
    case IDD_HEX9:
    case IDD_HEX10:
    case IDD_HEX11:
    case IDD_HEX12:
    case IDD_HEX13:
    case IDD_HEX14:
    case IDD_HEX15:
      idx = LOWORD(wParam) - IDD_HEX1;
      hexmode = (IsDlgButtonChecked(hwnd, LOWORD(wParam)) == BST_CHECKED);
      EnableWindow(GetDlgItem(hwnd, IDD_CASE1 + idx), !hexmode);
      return TRUE;
    } /* switch */
    break;

  } /* switch */


  return FALSE;
}

#pragma argsused
BOOL CALLBACK FilterInfoDlg(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
static
#include "autoreply-help.c"

  switch (msg) {
  case WM_INITDIALOG:
    SetDlgItemText(hwnd, INFOTEXT, helpstring);
    return TRUE;

  case WM_COMMAND:
    switch (LOWORD(wParam)) {
    case IDCANCEL:
    case IDOK:
      EndDialog(hwnd, LOWORD(wParam));
      return TRUE;
    } /* switch */
  } /* switch */

  return FALSE;
}

LRESULT FAR PASCAL SubClassFunc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
  switch (msg) {
  case WM_COMMAND:
    if (LOWORD(wParam) == ID_HELP && HIWORD(wParam) == BN_CLICKED)
      DialogBox(hinstDLL, "FilterInfo", hwnd, (DLGPROC)FilterInfoDlg);
    else if ((LOWORD(wParam) >= IDD_VAR1 && LOWORD(wParam) < IDD_VAR1 + NUM_VARIABLES) && HIWORD(wParam) == EN_CHANGE)
      reload(FALSE);
    break;

  default:
    return CallWindowProc(lpfnOldWndProc, hwnd, msg, wParam, lParam);
  } /* switch */

  return 0L;
}


static void reload(BOOL clearvars)
{
static int RecurseLevel = 0;
  int idx, seq, vpos, vlen;
  char key[32], value[MAX_EXPR_LEN + 16];
  char varlabel[MAX_VARLABEL_LEN], defvalue[MAX_VARDATA_LEN];
  char *ptr;

  /* avoid recursive loop due to this function changing edit fields, and a
     change in the edit fields causing a reload of the expressions */
  if (RecurseLevel > 0)
    return;
  RecurseLevel += 1;

  /* first clear all */
  for (idx = 0; idx < NUM_EXPRESSIONS; idx++) {
    Expression[idx].match[0] = '\0';
    Expression[idx].reply[0] = '\0';
    Expression[idx].match_len = 0;
    Expression[idx].reply_len = 0;
    Expression[idx].casesensitive = TRUE;
    Expression[idx].hexmode = FALSE;
  }
  if (clearvars) {
    for (idx = 0; idx < NUM_VARIABLES; idx++) {
      if (Variable[idx].hwndLabel) {
        DestroyWindow(Variable[idx].hwndLabel);
        Variable[idx].hwndLabel = NULL;
      }
      if (Variable[idx].hwndEdit) {
        DestroyWindow(Variable[idx].hwndEdit);
        Variable[idx].hwndEdit = NULL;
      }
      Variable[idx].expr_id = 0;
      Variable[idx].seq = 0;
      Variable[idx].label[0] = '\0';
      Variable[idx].data[0] = '\0';
    }
  }

  /* read settings (from the same INI file as Termite uses) */
  for (idx = 0; idx < NUM_EXPRESSIONS; idx++) {
    wsprintf(key, strExprKey, idx + 1);
    GetPrivateProfileString(strSection, key, "", value, sizearray(value), ProfileFile);
    ptr = strchr(value, '/');
    if (ptr != NULL && strlen(ptr + 1) < MAX_EXPR_LEN) {
      /* parse expression */
      *ptr++ = '\0';  /* skip '/' */
      Expression[idx].casesensitive = (strchr(value, 'i') == NULL);
      Expression[idx].hexmode = (strchr(value, 'x') != NULL);
      if (Expression[idx].hexmode) {
        Expression[idx].match_len = string2hex(ptr, Expression[idx].match, MAX_EXPR_LEN);
      } else {
        strcpy(Expression[idx].match, ptr);
        /* for the match expression, insert a backslash before every '(' and ')' */
        ptr = Expression[idx].match;
        while ((ptr = strpbrk(ptr, "()")) != NULL && strlen(Expression[idx].match) < MAX_EXPR_LEN - 1) {
          /* first check whether there is a backslash already */
          if (ptr != Expression[idx].match && *(ptr - 1) != '\\') {
            memmove(ptr + 1, ptr, strlen(ptr) + 1);
            *ptr = '\\';
            ptr += 2; /* skip both \ and the parenthesis */
          }
        }
        Expression[idx].match_len = strlen(Expression[idx].match);
      }
      /* parse reply */
      wsprintf(key, strReplyKey, idx + 1);
      GetPrivateProfileString(strSection, key, "", value, sizearray(value), ProfileFile);
      /* extract variables */
      seq = 0;
      while (parse_var(value, varlabel, defvalue, &vpos, &vlen)) {
        int v;
        for (v = 0; v < NUM_VARIABLES; v++)
          if (Variable[v].expr_id == idx && Variable[v].seq == seq && Variable[v].hwndEdit)
            break;
        if (v >= NUM_VARIABLES) {
          /* no variable found, find an empty spot (to add it) */
          for (v = 0; v < NUM_VARIABLES && Variable[v].hwndEdit; v++)
            /* nothing */;
          if (v < NUM_VARIABLES) {
            Variable[v].expr_id = idx;
            Variable[v].seq = seq;
            strcpy(Variable[v].label, varlabel);
            strcpy(Variable[v].data, defvalue);
            if (hwndPlugin) {
              Variable[v].hwndLabel = CreateWindow("static", Variable[v].label, WS_CHILD | WS_VISIBLE | SS_SIMPLE, (v*90)+90, 1, 40, 14, hwndPlugin, (HMENU)(-1), hinstDLL, NULL);
              Variable[v].hwndEdit = CreateWindow("edit", Variable[v].data, WS_CHILD | WS_VISIBLE | SS_SIMPLE, (v*90)+130, 0, 40, 16, hwndPlugin, (HMENU)(IDD_VAR1 + v), hinstDLL, NULL);
              SendMessage(Variable[v].hwndLabel, WM_SETFONT, (WPARAM)hfontNormal, MAKELPARAM(TRUE,0));
              SendMessage(Variable[v].hwndEdit, WM_SETFONT, (WPARAM)hfontNormal, MAKELPARAM(TRUE,0));
            }
          }
        } else {
          /* copy from the edit control to the data */
          assert(Variable[v].hwndEdit);
          GetWindowText(Variable[v].hwndEdit, Variable[v].data, MAX_VARDATA_LEN);
        }
        /* replace the variable by the value */
        strdel(value + vpos, vlen);
        strins(value + vpos, Variable[v].data, strlen(Variable[v].data));
        seq++;
      }
      if (Expression[idx].hexmode) {
        Expression[idx].reply_len = string2hex(value, Expression[idx].reply, MAX_EXPR_LEN);
      } else {
        strparse(Expression[idx].reply, sizearray(Expression[idx].reply), value);
        Expression[idx].reply_len = strlen(Expression[idx].reply);
      }
    }
  }

  RecurseLevel -= 1;
}

static BOOL parse_var(const char *string, char *label, char *defvalue,
                      int *vpos, int *vlen)
{
  struct slre_cap caps[2];
  int scanned;

  scanned = slre_match("{([A-Za-z0-9]+)=([^}]+)}", string, strlen(string), caps, 2);
  if (scanned <= 0
      || caps[0].ptr == NULL || caps[0].len >= MAX_VARLABEL_LEN
      || caps[1].ptr == NULL || caps[1].len >= MAX_VARDATA_LEN)
        return FALSE;
  assert(label != NULL);
  assert(caps[0].len > 0);
  strncpy(label, caps[0].ptr, caps[0].len);
  label[caps[0].len] = '\0';

  assert(defvalue != NULL);
  assert(caps[1].len > 0);
  strncpy(defvalue, caps[1].ptr, caps[1].len);
  defvalue[caps[1].len] = '\0';

  assert(vpos != NULL);
  *vpos = (int)(caps[0].ptr - string) - 1;  /* -1 because { comes just before caps[0] */
  assert(vlen != NULL);
  *vlen = caps[0].len + caps[1].len + 3;    /* +3 for { = } */
  return TRUE;
}

static int match_re(const char *string, int length, int *scanned)
{
  struct slre_cap caps[2];
  char expr[MAX_EXPR_LEN+10]; /* for the parentheses around the expression and the (?i) option */
  int idx;

  for (idx = 0; idx < NUM_EXPRESSIONS; idx += 1) {
    if (Expression[idx].match_len == 0)
      continue;         /* an empty string matches nothing */
    if (Expression[idx].hexmode)
      continue;         /* skip hex-mode strings (handled separately) */

    if (Expression[idx].casesensitive)
      strcpy(expr, "(?i)(");
    else
      strcpy(expr, "(");
    strcat(expr, Expression[idx].match);
    strcat(expr, ")");
    assert(scanned != NULL);
    *scanned = slre_match(expr, string, length, caps, 2);
    if (*scanned >= 0)
      return idx;
  }
  return -1;
}

static int match_hex(const char *string, int length, int *scanned)
{
  int idx;
  const char *ptr;

  for (idx = 0; idx < NUM_EXPRESSIONS; idx += 1) {
    if (Expression[idx].match_len == 0)
      continue;         /* an empty string matches nothing */
    if (!Expression[idx].hexmode)
      continue;         /* skip regex strings (handled separately) */

    ptr = string;
    for ( ;; ) {
      /* find first byte */
      while ((ptr - string) < length && *ptr != Expression[idx].match[0])
        ptr++;
      if ((ptr - string) >= length)
        break;  /* first byte not found, quit loop */
      /* compare the remaining bytes */
      if (Expression[idx].match_len == 1 || memcmp(ptr, Expression[idx].match, Expression[idx].match_len) == 0) {
        *scanned = (ptr - string) + Expression[idx].match_len;
        return idx;
      }
    }
  }
  return -1;
}

static int hexchar(char ch)
{
  if ('0' <= ch && ch <='9')
    return ch - '0';
  if ('a' <= ch && ch <='f')
    return ch - 'a' + 10;
  if ('A' <= ch && ch <='F')
    return ch - 'A' + 10;
  return -1;
}

/* strparse() copies a string from input to output and translates escaped
 * characters to the original characters. The return value is whether all escape
 * sequences are valid, and if the complete string was copied. If there are one
 * or more invalid escape sequences, the returned result is FALSE.
 */
static BOOL strparse(char *output, int size, const char *input)
{
  BOOL result = TRUE;
  int count = 0;

  assert(output != NULL);
  assert(size > 0);
  assert(input != NULL);
  while (*input != '\0' && count < size) {
    if (*input == '\\') {
      const char *skip = input + 1;
      char byte = 0;
      switch (*skip) {
      case 'n':
        byte = '\n';
        break;
      case 'r':
        byte = '\r';
        break;
      case 't':
        byte = '\t';
        skip++;
        break;
      case 'x':
        if (hexchar(skip[1]) >= 0 && hexchar(skip[2]) >= 0) {
          byte = (hexchar(skip[1]) << 4) | hexchar(skip[2]);
          skip += 2;
        }
        break;
      case '\\':
        byte = '\\';
        break;
      }
      if (byte != 0) {
        *output = byte;
        input = skip;
      } else {
        *output = *input;
        result = FALSE;
      }
    } else {
      *output = *input;
    }
    input++;
    output++;
    count++;
  }

  /* zero-terminate */
  if (count < size)
    *output = '\0';
  else
    result = FALSE;

  return result;
}

static int string2hex(const char *string, unsigned char *buffer, int buflen)
{
  int idx = 0;
  char *ptr = (char*)string;
  for ( ;; ) {
    if (idx >= buflen)
      break;
    while (*ptr != '\0' && *ptr <= ' ')
      ptr++;
    if (*ptr == '\0')
      break;
    buffer[idx] = strtol(ptr, &ptr, 16);
    idx++;
  }

  return idx;
}

static char *strdel(char *str,size_t len)
{
  size_t length=strlen(str);
  if (len>length)
    len=length;
  memmove(str,str+len,length-len+1);  /* include EOS byte */
  return str;
}

static char *strins(char *dest,const char *src,size_t srclen)
{
  size_t destlen=strlen(dest);
  memmove(dest+srclen,dest,destlen+1);/* include EOS byte */
  memcpy(dest,src,srclen);
  return dest;
}

